/* condvar.c */

/* Sort of condition variables, similar to locks but not the same
 * (see ../sem/lock.c) 
 */

/* Implementa (singolo) condvar; il suo stato e`:
- coda processi bloccati
- intero PENDING: |PENDING| = n. processi in coda;
  se < 0 i processi sono bloccati su SIGNAL, se > 0 su WAIT
                
                PENDING = 0        PENDING < 0     PENDING > 0
 
      WAIT()    process blocks     rendez-vous     process blocks
                PENDING++          PENDING++       PENDING++

    SIGNAL()    process blocks     process blocks  rendez-vous
                PENDING--          PENDING--       PENDING--
      
*/
 
/* NB: SIGNAL()'s block and do not accumulate (like up's on semaphores). 
 * Con le vere condition variables signal() non dovrebbe forse bloccare;
 * il vantaggio di avere signal() bloccanti e` di costringere il 
 * programmatore a sincronizzarle con i WAIT().
 */ 

#include <stdio.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

extern int sem_id;

struct sembuf op_SIGNAL[2] = {	// Blocking (unusual)!
  { 0, 0, 0 },					/* aspetta che il semaforo 0 diventi 0 */
  { 0, 1, SEM_UNDO }			/* poi incrementa di 1 il semaforo 0 */
} ;
int SIGNAL(int sem_num)
{
  op_SIGNAL[0].sem_num = sem_num;
  op_SIGNAL[1].sem_num = sem_num;
  if (semop(sem_id, &op_SIGNAL[0], 2) < 0) {
    perror("signal del semaforo");
    exit(-1);
  }
  return 0;
}

struct sembuf op_WAIT[1] = {// Blocking
  {0, -1, SEM_UNDO}			/* decrementa il semaforo 0 di 1 */
} ;
int WAIT(int sem_num)
{
  op_WAIT[0].sem_num = sem_num;
  if (semop(sem_id,&op_WAIT[0],1) < 0) {
    perror("wait del semaforo");
    exit(-1);
  }
  return 0;
}